import numpy as np
from hashlib import sha256
import json


class NamingMixin:
    """ """
    @property
    def __name__(self):
        try:
            ctx_short_name = self.short_name
        except:
            ctx_short_name = self.get_short_name()
        try:
            ctx_tracked_params = self.tracked_params
        except:
            ctx_tracked_params = self.get_tracked_params()
        self.params = {
            key: self.get_params()[key]
            for key in ctx_tracked_params if key in self.get_params()
        }
        return f"{ctx_short_name}_{sha256((json.dumps(self.params, sort_keys=True)).encode()).hexdigest()[-3:] if self.params else 'default'}"


class ArithmeticMixin:
    """ """
    def __mul__(self, x):
        return WeightedModel(self, x)

    def __rmul__(self, x):
        return self * x

    def __truediv__(self, x):
        assert x != 0
        return WeightedModel(self, 1.0 / x)

    def __add__(self, other):
        # assert type(self).__base__ == type(other).__base__, \
        # f"Can't add {type(other).__name__} to {type(self).__name__}."
        if other == 0:
            return self
        return Ensemble([self, other])

    def __radd__(self, other):
        return self + other


class SourceMixin:
    """ """
    @property
    def source(self):
        """ """
        args = []
        for key, value in self.get_params().items():
            args.append(f"{key}={repr(value)}")
        res = ", ".join(args)
        return f"{self.__class__.__name__}({res})"


class PreprocessingMixin:
    """ """
    def preprocess(self, X, y):
        """Preprocess input before feeding it into model

        Args:
          X: np.array
          y: np.array or None (fitting or inference)

        Returns:
          X_processed, y_processed)

        """
        return X, y

    def preprocess_fit(self, X, y, *args, **kwargs):
        """

        Args:
          X:
          y:
          *args:
          **kwargs:

        Returns:

        """
        X_proc, y_proc = self.preprocess(X, y)
        self.fit(X_proc, y_proc, *args, **kwargs)

    def preprocess_predict(self, X, *args, **kwargs):
        """

        Args:
          X:
          *args:
          **kwargs:

        Returns:

        """
        X_proc, _ = self.preprocess(X, None)
        return self.predict(X_proc, *args, **kwargs)


class Model(ArithmeticMixin, NamingMixin, SourceMixin, PreprocessingMixin):
    """ """
    def __repr__(self):
        return f"[{self.__name__}] {self.source}"


class WeightedModel(ArithmeticMixin, PreprocessingMixin):  # MulNode
    """Multiplies predictions by a certain coefficient."""
    def __init__(self, model, coeff):
        self.model = model
        self.coeff = coeff
        if type(self.model) == type(self):
            self.coeff *= self.model.coeff
            self.model = self.model.model
        self.__name__ = f"{round(self.coeff, 2)} * "
        if "+" in self.model.__name__:
            self.__name__ += f"({self.model.__name__})"
        else:
            self.__name__ += f"{self.model.__name__}"

    def predict(self, X):
        """Standard prediction method.

        Args:
          X: data matrix

        Returns:

        """
        return self.coeff * self.model.preprocess_predict(X)


class Ensemble(ArithmeticMixin, PreprocessingMixin):  # AddNode
    """Sums up all predictions of all models."""
    def __init__(self, models):
        self.models = []
        ensembles = [model for model in models if type(model) == type(self)]
        single_models = [model for model in models if type(model) != type(self)]
        _models = []
        for ensemble in ensembles:
            _models += ensemble.models
        _models += single_models

        def __get_model_coeff(model):
            """
            :param model:
            :return: if the model is weighted, then returns its coefficient, else 1
            """
            if isinstance(model, WeightedModel):
                return model.model, model.coeff
            return model, 1

        self.models = list(
            set([__get_model_coeff(model)[0] for model in _models]))
        _coeffs = np.zeros(len(self.models))
        for i in range(len(self.models)):
            for model in _models:
                if __get_model_coeff(
                        self.models[i])[0] == __get_model_coeff(model)[0]:
                    _coeffs[i] += __get_model_coeff(model)[1]

        self.models = [
            WeightedModel(model, coeff)
            for model, coeff in zip(self.models, _coeffs)
        ]
        self.norm_coeff = _coeffs.sum()
        _coeffs /= self.norm_coeff
        self.__name__ = " + ".join([(model / self.norm_coeff).__name__
                                    for model in self.models])

    def predict(self, X):
        """Standard prediction method.

        Args:
          X: data matrix

        Returns:

        """
        res = 0
        for model in self.models:
            res += model.preprocess_predict(X)
        return res / self.norm_coeff

    def __mul__(self, x):
        return Ensemble([model * x for model in self.models])

    def __truediv__(self, x):
        assert x != 0
        return Ensemble([model / x for model in self.models])


class BinaryClassifierMixin(Model):
    """ """
    def predict(self, X, **kwargs):
        """

        Args:
          X:
          **kwargs:

        Returns:

        """
        return self.predict_proba(X, **kwargs)[:, 1]


class MultiClassifierMixin(Model):
    """ """
    def predict(self, X, **kwargs):
        """

        Args:
          X:
          **kwargs:

        Returns:

        """
        return self.predict_proba(X, **kwargs)


class RegressorMixin(Model):
    """ """

    pass



# import os
#
# from . import config
# from .feature.helper import helper_list
# from .utils import captcha
#
#
# class Submission:
#     """ """
#     def __init__(self, exp):
#         self.exp = exp
#         self.fs = exp.featureset
#         self.model = exp.model
#         self.metric = exp.metric
#         self.splitter = exp.splitter
#         self.features = exp.features
#         self.helpers = exp.helpers
#         self.path = config.ROOT_DIR + f'submissions/{exp.identifier}/'
#
#     def feature_source(self):
#         """FEATURE_SOURCE_TEMPLATE = f"""\
#         import kts
#         from kts import *
#
#         Args:
#
#         Returns:
#         return FEATURE_SOURCE_TEMPLATE + '\n\n'.join(['@register\n' + f.source for f in self.features])
#
#     def helper_source(self):
#         """HELPER_SOURCE_TEMPLATE = f"""\
#         import kts
#         from kts import *
#
#         Args:
#
#         Returns:
#         return HELPER_SOURCE_TEMPLATE + '\n\n'.join(['@helper\n' + f.source for f in self.helpers])
#
#     def main_source(self):
#         """MAIN_SOURCE_TEMPLATE = f"""
#         import kts
#         from kts import *
#
#         from features import *
#         from helpers import *
#
#         from {self.splitter.__module__} import {self.splitter.__class__.__name__}
#         {f"from {self.metric.__module__} import {self.metric.__name__}" if self.metric.__module__ != '__main__' else ""}
#         from {self.model.__module__} import {self.model.__class__.__name__}
#
#
#         splitter = {self.splitter}
#         metric = {self.metric.__name__}
#         val = Validator(splitter, metric)
#
#         {self.fs.__name__} = {self.fs.source}
#
#         {self.model.__name__} = {self.model.source}
#
#         val.score({self.model.__name__}, {self.fs.__name__})
#
#         Args:
#
#         Returns:
#         return MAIN_SOURCE_TEMPLATE
#
#     def write(self):
#         """ """
#         if not os.path.exists(self.path):
#             os.mkdir(self.path)
#             os.system(f'cd {self.path} && y | kts init')
#         else:
#             print('Such submission already exists. Rewrite?')
#             if not captcha():
#                 return
#         with open(self.path + 'features.py', 'w') as f:
#             f.write(self.feature_source())
#         with open(self.path + 'helpers.py', 'w') as f:
#             f.write(self.helper_source())
#         with open(self.path + 'main.py', 'w') as f:
#             f.write(self.main_source())
#         try:
#             create_submission = helper_list['create_submission']
#         except:
#             raise UserWarning("You should create helper with name create_submission(experiment) -> pd.DF")
#         df_sub = create_submission(self.exp)
#         df_sub.to_csv(self.path + 'submission.csv')
#
#
# def submit(experiment):
#     """
#
#     Args:
#       experiment:
#
#     Returns:
#
#     """
#     subm = Submission(experiment)
#     subm.write()



import hashlib
import inspect
import time
from abc import ABCMeta
from itertools import zip_longest

import numpy as np


def captcha():
    """ """
    np.random.seed(int(time.time()))
    a, b = np.random.randint(5, 30, size=2)
    c = int(input(f"{a} + {b} = "))
    if a + b != c:
        return False
    return True


def list_hash(lst, length):
    """

    Args:
      lst: 
      length: 

    Returns:

    """
    return hashlib.sha256(repr(tuple(lst)).encode()).hexdigest()[:length]


def hash_str(a):
    """

    Args:
      a: 

    Returns:

    """
    return hashlib.sha256(a.encode()).hexdigest()


def extract_signature(func, return_dict=False):
    """

    Args:
      func: 
      return_dict:  (Default value = False)

    Returns:

    """
    args = inspect.getfullargspec(func).args
    defaults = inspect.getfullargspec(func).defaults
    values = {
        **inspect.currentframe().f_back.f_locals,
        **inspect.currentframe().f_back.f_back.f_locals,
    }
    if defaults is None:
        defaults = []
    if args is None:
        args = []
    # print(inspect.getfullargspec(func))
    # print(values)
    if return_dict:
        return {arg: values[arg] for arg in args}
    sources = []
    for arg, default in list(zip_longest(args[::-1], defaults[::-1]))[::-1]:
        if values[arg] != default:
            if is_helper(values[arg]):
                arg_repr = values[arg].__name__
            else:
                arg_repr = repr(values[arg])
            sources.append(f"{arg}={arg_repr}")
    return ", ".join(sources)


def is_helper(func):
    """

    Args:
      func: 

    Returns:

    """
    return (callable(func) and "__name__" in dir(func)
            and "source" in dir(func) and isinstance(func.source, str)
            and "def" in func.source)  # genius


class BadSignatureException(Exception):
    """ """

    pass


def _create_source(class_name, base_classes, methods):
    """

    Args:
      class_name: 
      base_classes: 
      methods: 

    Returns:

    """
    base_class_names = ", ".join([bc.__name__ for bc in base_classes])
    res = f"""class {class_name}({base_class_names}):\n"""
    for name, meth in methods.items():
        if name not in ["__module__", "__qualname__", "__doc__"]:
            try:
                res += inspect.getsource(meth) + "\n"
            except TypeError:
                raise UserWarning(
                    f"Unexpected static variable in class definition: {name}. Use instance variables instead."
                )
    return res.strip()


def _check_signatures(class_name, base_classes, methods):
    """

    Args:
      class_name: 
      base_classes: 
      methods: 

    Returns:

    """
    for name, meth in methods.items():
        for base_class in base_classes:
            if name in base_class.__dict__ and name not in [
                "__module__",
                "__qualname__",
                "__doc__",
            ]:
                base_meth = getattr(base_class, name)
                sign = inspect.signature(meth)
                base_sign = inspect.signature(base_meth)
                if sign != base_sign:
                    raise BadSignatureException(
                        f"Signature discrepancy: {name}{sign} in {class_name}, "
                        f"but {name}{base_sign} in {base_class.__name__}")


class SourceMetaClass(ABCMeta):
    """ """
    def __new__(cls, class_name, base_classes, dict_of_methods):
        cls.check_methods(dict_of_methods)
        _check_signatures(class_name, base_classes, dict_of_methods)
        dict_of_methods["class_source"] = _create_source(
            class_name, base_classes, dict_of_methods)
        return type.__new__(cls, class_name, base_classes, dict_of_methods)

    def check_methods(methods):
        """

        Args:
          methods: 

        Returns:

        """
        pass



import seaborn as sns
from matplotlib.pyplot import figure

from kts.feature_selection.selector import BuiltinImportance


def plot_importances(experiment,
                     n_best=15,
                     sort_by="max",
                     calculator=BuiltinImportance(),
                     fontsize=12):
    """Visualize feature importances (max, mean and std) of an experiment using a given calculator.

    Args:
      experiment: Experiment instance, like lb['012ABC']
      sort_by: one of 'max', 'mean' and 'std' (Default value = 'max')
      calculator: ImportanceCalculator instance (Default value = BuiltinImportance())
      fontsize: return: (Default value = 12)
      n_best:  (Default value = 15)

    Returns:

    """
    assert sort_by in ["max", "mean", "std"]
    importances = experiment.feature_importances(
        importance_calculator=calculator)
    tmp = importances.agg(["max", "mean", "std"])
    tmp = tmp.T
    tmp = tmp.reset_index()
    tmp.sort_values(sort_by, ascending=False, inplace=True)
    figure(figsize=(8, n_best / 3))
    sns.barplot(x="max", y="index", data=tmp.head(n_best), alpha=0.5)
    sns.barplot(x="mean", y="index", data=tmp.head(n_best), alpha=0.7)
    a = sns.barplot(x="std", y="index", data=tmp.head(n_best), alpha=1)
    a.tick_params(labelsize=fontsize)
    a.set_xlabel("importance", fontsize=fontsize)
    a.set_ylabel("feature", fontsize=fontsize)
    return a



from sklearn.model_selection import BaseCrossValidator
from sklearn.utils import indexable


class Refiner(BaseCrossValidator):
    """ """
    def __init__(self, outer_splitter, inner_splitter=None):
        super().__init__()
        self.outer_splitter = outer_splitter
        self.inner_splitter = (inner_splitter if inner_splitter is not None
                               else outer_splitter)

    def split(self, X, y=None, groups=None):
        """

        Args:
          X: 
          y:  (Default value = None)
          groups:  (Default value = None)

        Returns:

        """
        X, y, groups = indexable(X, y, groups)
        for _, idx_test_outer in self.outer_splitter.split(X, y, groups):
            for idx_train, idx_test in self.inner_splitter.split(
                    X[idx_test_outer],
                    y[idx_test_outer] if y is not None else None,
                    groups[idx_test_outer] if groups is not None else None,
            ):
                yield idx_test_outer[idx_train], idx_test_outer[idx_test]

    def get_n_splits(self, X=None, y=None, groups=None):
        """

        Args:
          X:  (Default value = None)
          y:  (Default value = None)
          groups:  (Default value = None)

        Returns:

        """
        return self.outer_splitter.get_n_splits() * self.inner_splitter.get_n_splits()



import pandas as pd

from kts.core.dataframe import DataFrame as KTDF
import kts.core.feature_constructor
from kts.util.misc import extract_signature


def wrap_stl_function(outer_function, inner_function):
    """

    Args:
      outer_function:
      inner_function:

    Returns:

    """
    source = f"stl.{outer_function.__name__}({extract_signature(outer_function)})"
    fc = kts.core.feature_constructor.FeatureConstructor(inner_function, cache_default=False, stl=True)
    fc.source = source
    fc.args = extract_signature(outer_function, return_dict=True)
    return fc


def __empty_like(df):
    if isinstance(df, KTDF):
        return KTDF(df[[]].copy(), df.train, df.encoders, df.slice_id)
    else:
        return KTDF(df[[]].copy())


empty_like = kts.core.feature_constructor.FeatureConstructor(__empty_like, cache_default=False)
empty_like.stl = True
empty_like.source = "stl.empty_like"

identity = kts.core.feature_constructor.FeatureConstructor(lambda df: df, cache_default=False)
identity.stl = True
identity.source = "stl.identity"


def merge(dfs):
    """

    Args:
      dfs:

    Returns:

    """
    if len(dfs) == 1:
        return dfs[0]
    return pd.concat(dfs, axis=1)


def column_selector(columns):
    """

    Args:
      columns:

    Returns:

    """
    def __col_selector(df):
        return df[[col for col in columns if col in df.columns]]

    # return FeatureConstructor(__col_selector, cache_default=False)
    return wrap_stl_function(column_selector, __col_selector)


def column_dropper(columns):
    """

    Args:
      columns:

    Returns:

    """
    def __col_dropper(df):
        return df.drop([col for col in columns if col in df.columns], axis=1)

    # return FeatureConstructor(__col_dropper, cache_default=False)
    return wrap_stl_function(column_dropper, __col_dropper)


def concat(funcs):
    """

    Args:
      funcs:

    Returns:

    """
    funcs = sum(
        [
            i.args["funcs"] if i.source.startswith("stl.concat") else [i]
            for i in funcs
        ],
        [],
    )

    def __concat(df):
        return merge([func(df) for func in funcs])

    fc = wrap_stl_function(concat, __concat)
    fc.source = f"stl.concat([{', '.join([func.__name__ if not func.stl else func.source for func in funcs])}])"
    return fc


def compose(funcs):
    """

    Args:
      funcs:

    Returns:

    """
    if (len(funcs) == 2 and funcs[0].source.startswith("stl.concat")
            and funcs[1].source.startswith("stl.column_selector")):
        selection_list = funcs[1].args["columns"]
        return concat(
            [func + selection_list for func in funcs[0].args["funcs"]])

    def __compose(df):
        res = identity(df)
        for func in funcs:
            res = func(res)
        return res

    fc = wrap_stl_function(compose, __compose)
    fc.source = f"stl.compose([{', '.join([func.__name__ if not func.stl else func.source for func in funcs])}])"
    return fc


import re
import warnings

from kts.core.backend.memory import cache


def is_identifier(string):
    """

    Args:
      string: 

    Returns:

    """
    return isinstance(string, str) and bool(re.match("[0-9A-F]{6}", string))


def is_list_of_identifiers(a):
    """

    Args:
      a: 

    Returns:

    """
    if not isinstance(a, list):
        return False
    for i in a:
        if not is_identifier(i):
            return False
    return True


def get_experiment(identifier):
    """

    Args:
      identifier: 

    Returns:

    """
    assert is_identifier(identifier), "You can only get experiment by ID"
    names = [
        name for name in cache.cached_objs()
        if name.endswith("_exp") and name.startswith(identifier)
    ]
    if len(names) == 0:
        raise KeyError("No such experiment found")
    elif len(names) == 1:
        return cache.load_obj(names[0])
    else:
        warnings.warn(
            "I don't know what happened but there are several objects with this ID"
        )
        return [cache.load_obj(name) for name in names]



from fastprogress import progress_bar

from kts.modelling.mixins import ArithmeticMixin, PreprocessingMixin


class Pipeline(ArithmeticMixin, PreprocessingMixin):
    """ """
    def __init__(self, model, featureslice):
        self.model = model
        self.featureslice = featureslice
        self.__name__ = (self.model.__name__ + "-" +
                         self.featureslice.featureset.__name__)

    def fit(self, **kwargs):
        """

        Args:
          **kwargs: 

        Returns:

        """
        if "masterbar" in kwargs:
            mb = kwargs.pop("masterbar")
        else:
            mb = None
        if mb:
            pb = progress_bar(range(2), parent=mb)
            mb.child.comment = f"computing the features..."
            pb.on_iter_begin()
            pb.update(0)
        X = self.featureslice().values
        y = self.featureslice.target.values
        if mb:
            mb.child.comment = f"training..."
            pb.on_iter_begin()
            pb.update(1)
        try:
            self.model.preprocess_fit(
                X,
                y,
                eval_set=[(
                    self.featureslice(self.featureslice.idx_test).values,
                    self.featureslice.target.values[
                        self.featureslice.idx_test],
                )],
                **kwargs,
            )
        except:
            if mb:
                mb.child.comment = (
                    "failed to train with eval_set, training without it...")
            self.model.preprocess_fit(X, y, **kwargs)
        if mb:
            pb.update(2)

    def predict(self, df, **kwargs):
        """

        Args:
          df: 
          **kwargs: 

        Returns:

        """
        if "masterbar" in kwargs:
            mb = kwargs.pop("masterbar")
        else:
            mb = None
        if mb:
            pb = progress_bar(range(2), parent=mb)
            mb.child.comment = f"computing the features..."
            pb.on_iter_begin()
            pb.update(0)
        X = self.featureslice(df).values
        if mb:
            mb.child.comment = f"predicting..."
            pb.on_iter_begin()
            pb.update(1)
        res = self.model.preprocess_predict(X, **kwargs)
        if mb:
            pb.update(2)
        return res



from kts import config
from kts.core import dataframe
from kts.core.backend import memory
from kts.util import source_utils, cache_utils
from kts.util.misc import captcha


class FeatureConstructor:
    """ """
    def __init__(self, function, cache_default=True, stl=False):
        self.function = function
        self.cache_default = cache_default
        self.__name__ = function.__name__
        self.source = source_utils.get_source(function)
        self.stl = stl
        self.args = dict()

    # needs refactoring because of direct storing source
    def __call__(self, df, cache=None, **kwargs):
        if not self.stl:
            self = memory.cache.load_obj(self.__name__ + "_fc")
        ktdf = dataframe.DataFrame(df=df)
        if type(cache) == type(None):
            cache = self.cache_default
        if not cache or config.IS_PREVIEW_CALL:  # written to avoid caching when @preview function uses @registered function inside
            return self.function(ktdf, **kwargs)

        name = f"{self.function.__name__}__{cache_utils.get_hash_df(ktdf)[:4]}__{ktdf.slice_id[-4:]}"
        name_metadata = name + "_meta"
        if memory.cache.is_cached_df(name):
            if memory.cache.is_cached_obj(name_metadata):
                cached_encoders = memory.cache.load_obj(name_metadata)
                for key, value in cached_encoders.items():
                    ktdf.encoders[key] = value
            return dataframe.DataFrame(
                df=memory.cache.load_df(name),
                train=ktdf.train,
                encoders=ktdf.encoders,
                slice_id=ktdf.slice_id,
            )
        else:
            result = self.function(ktdf)
            try:
                memory.cache.cache_df(result, name)
            except MemoryError:
                print(f"The dataframe is too large to be cached. "
                      f"It is {cache_utils.get_df_volume(result)}, "
                      f"but current memory limit is {config.MEMORY_LIMIT}.")
                print(f"Setting memory limit to {cache_utils.get_df_volume(result) * 2}")
                ans = input('Please confirm memory limit change. Enter "No" to cancel it.')
                do_change = True
                if ans.lower() == "no":
                    if captcha():
                        do_change = False
                if do_change:
                    memory.cache.set_memory_limit(
                        cache_utils.get_df_volume(result) * 2)
                    memory.cache.cache_df(result, name)
                    print("Done. Please don't forget to update your kts_config.py file.")
                else:
                    print("Cancelled")
            if ktdf.encoders:
                memory.cache.cache_obj(ktdf.encoders, name_metadata)
            return dataframe.DataFrame(
                df=result,
                train=ktdf.train,
                encoders=ktdf.encoders,
                slice_id=ktdf.slice_id,
            )

    def __repr__(self):
        if self.stl:
            return self.source
        else:
            return f'<Feature Constructor "{self.__name__}">'

    def __str__(self):
        return self.__name__

    def __sub__(self, other):
        assert isinstance(other, list)
        import kts.core.base_constructors

        return kts.core.base_constructors.compose([self, kts.core.base_constructors.column_dropper(other)])

    def __add__(self, other):
        assert isinstance(other, list)
        import kts.core.base_constructors

        return kts.core.base_constructors.compose([self, kts.core.base_constructors.column_selector(other)])

    def __mul__(self, other):
        assert isinstance(other, FeatureConstructor)
        import kts.core.base_constructors

        return kts.core.base_constructors.compose([self, other])


import multiprocessing

import numpy as np
import pandas as pd

from kts.validation.leaderboard import leaderboard as lb
from kts.core.backend.memory import load
from kts.core.base_constructors import merge, wrap_stl_function, empty_like


def _apply_df(args):
    """

    Args:
      args:

    Returns:

    """
    df, func, num, kw = args
    return num, df.swifter.apply(func, **kw)


def apply(dataframe, function, **kwargs):
    """Applies function to dataframe faster.
    If n_threads is in kwargs and is greater than 1, applies by multiprocessing.
    :return: same as df.apply(function)

    Args:
      dataframe:
      function:
      **kwargs:

    Returns:

    """
    if "n_threads" in kwargs:
        n_threads = kwargs.pop("n_threads")
    else:
        n_threads = 1
    if n_threads == 1:
        return dataframe.swifter.apply(function, **kwargs)

    pool = multiprocessing.Pool(processes=n_threads)
    result = pool.map(
        _apply_df,
        [(d, function, i, kwargs)
         for i, d in enumerate(np.array_split(dataframe, n_threads))],
    )
    pool.close()
    result = sorted(result, key=lambda x: x[0])
    return pd.concat([i[1] for i in result])


def get_categorical(df):
    """

    Args:
      df:

    Returns:

    """
    cat_features = [col for col in df.columns if df[col].dtype == object]
    return cat_features


def get_numeric(df):
    """

    Args:
      df:

    Returns:

    """
    num_features = [col for col in df.columns if df[col].dtype != object]
    return num_features


def stack(ids):
    """

    Args:
      ids:

    Returns:

    """
    def __stack(df):
        oof_preds = merge([lb[id_exp].oof for id_exp in ids])
        oof_col_names = merge([lb[id_exp].oof.columns for id_exp in ids])
        try:
            res = oof_preds[df.index]
            if res.isna().sum().sum() > 0:
                bad_indices = res.isna().sum(axis=1) > 0
                preds = [lb[id_exp].predict(df[bad_indices]) for id_exp in ids]
                res[bad_indices] = merge([
                    pd.DataFrame(data=pred, columns=col_names)
                    for col_names, pred in zip(oof_col_names, preds)
                ])
        except:
            preds = [lb[id_exp].predict(df) for id_exp in ids]
            res = merge([
                pd.DataFrame(data=pred, columns=col_names)
                for col_names, pred in zip(oof_col_names, preds)
            ])
            res.set_index(df.index, inplace=True)
        return res

    return wrap_stl_function(stack, __stack)


def from_df(name):
    """

    Args:
      name:

    Returns:

    """
    def __from_df(df):
        tmp = load(name)
        return tmp.loc[df.index]

    return wrap_stl_function(from_df, __from_df)


def cv_apply(feature_constructor, split):
    """

    Args:
      feature_constructor:
      split:

    Returns:

    """
    raise NotImplementedError

    def __cv_apply(df):
        res = empty_like(df)


import abc
from abc import ABC

import numpy as np
from sklearn.metrics import make_scorer
from sklearn.model_selection._split import _build_repr


class ImportanceCalculator(ABC):
    """ """
    @property
    @abc.abstractmethod
    def short_name(self):
        """ """
        pass

    @abc.abstractmethod
    def calc(self, model, featureslice, experiment) -> dict:
        """

        Args:
          model: 
          featureslice: 
          experiment: 

        Returns:

        """
        raise NotImplementedError

    def __repr__(self):
        return _build_repr(self)


class BuiltinImportance(ImportanceCalculator):
    """ """

    short_name = "bltn"

    def __init__(self):
        super().__init__()
        pass

    def calc(self, model, featureslice, experiment):
        """

        Args:
          model: 
          featureslice: 
          experiment: 

        Returns:

        """
        return {
            name: imp for name, imp in zip(featureslice.columns, model.feature_importances_)
        }


# class LegacyImportance(ImportanceCalculator):
#     short_name = 'lgc'
#
#     def __init__(self, df):
#         super().__init__()
#         self.df = df
#
#     def calc(self, model, featureslice):
#         col_names = list(featureslice(df)[:5].columns)
#         return {name: imp for name, imp in zip(col_names, model.feature_importances_)}

try:
    import eli5.sklearn

    class SklearnPermutationImportance(ImportanceCalculator):
        """ """

        short_name = "perm"

        def __init__(self, n_rows=1000, n_iter=5, random_state=42):
            super().__init__()
            self.n_rows = n_rows
            self.n_iter = n_iter
            self.random_state = random_state

        def calc(self, model, featureslice, experiment):
            """

            Args:
              model: 
              featureslice: 
              experiment: 

            Returns:

            """
            if ("df_input" not in dir(featureslice.featureset)
                    or featureslice.featureset.df_input is None):
                raise AttributeError(
                    f"No input dataframe for featureset of the experiment found. "
                    f"Set it with lb['{experiment.identifier}'].set_df(df)")
            perm = eli5.sklearn.PermutationImportance(
                model,
                scoring=make_scorer(experiment.metric),
                refit=False,
                n_iter=self.n_iter,
                random_state=self.random_state,
            )
            perm.fit(
                featureslice(featureslice.idx_test[:self.n_rows]),
                featureslice.featureset.target.values[featureslice.idx_test][:self.n_rows],
            )
            return {
                name: imp
                for name, imp in zip(featureslice.columns, perm.feature_importances_)
            }

except ImportError:
    pass

try:
    from eli5.permutation_importance import get_score_importances

    class PermutationImportance(ImportanceCalculator):
        """ """

        short_name = "perm"

        def __init__(self, n_rows=1000, n_iter=5, random_state=42):
            super().__init__()
            self.n_rows = n_rows
            self.n_iter = n_iter
            self.random_state = random_state

        def calc(self, model, featureslice, experiment):
            """

            Args:
              model: 
              featureslice: 
              experiment: 

            Returns:

            """
            if ("df_input" not in dir(featureslice.featureset)
                    or featureslice.featureset.df_input is None):
                raise AttributeError(
                    f"No input dataframe for featureset of the experiment found. "
                    f"Set it with lb['{experiment.identifier}'].set_df(df)")

            def score_func(X, y):
                """

                Args:
                  X: 
                  y: 

                Returns:

                """
                return experiment.metric(y, model.predict(X))

            X = featureslice(featureslice.idx_test[:self.n_rows]).values
            y = featureslice.featureset.target.values[featureslice.idx_test][:self.n_rows]
            base_score, score_decreases = get_score_importances(
                score_func,
                X,
                y,
                n_iter=self.n_iter,
                random_state=self.random_state)
            feature_importances = np.mean(score_decreases, axis=0)
            return {
                name: imp
                for name, imp in zip(featureslice.columns, feature_importances)
            }

except ImportError:
    pass



from copy import deepcopy

import pandas as pd


class DataFrame(pd.DataFrame):
    """A wrapper over the standard DataFrame class.
    Complements it with .train and .encoders attributes.
    
    This class is implemented to provide an indicator
    for functions whether they serve a train or test call and let propagate
    this indicator further to inner functions.

    Example:
        ```
        def a(df):
            res = stl.empty_like(df)
            tmp = b(df)
            res['res'] = b['b'] ** 2
            return res
        ```
    Args:

    Returns:

    """

    _metadata = ["train", "encoders", "slice_id"]
    _accessors = frozenset(_metadata)

    def __init__(self, df, train=None, encoders=None, slice_id=None, **kwargs):
        if isinstance(df, DataFrame):
            pd.DataFrame.__init__(self, df, **kwargs)
            self.slice_id = (df.slice_id if isinstance(slice_id, type(None)) else slice_id)
            self.train = df.train if isinstance(train, type(None)) else train
            self.encoders = (df.encoders if isinstance(encoders, type(None)) else encoders)
        else:
            pd.DataFrame.__init__(self, df, **kwargs)
            self.slice_id = "0" * 16 if isinstance(slice_id, type(None)) else slice_id
            self.train = False if isinstance(train, type(None)) else train
            self.encoders = dict() if isinstance(encoders, type(None)) else encoders

    def __copy__(self, deep=False):
        if not deep:
            return DataFrame(
                df=self,
                slice_id=self.slice_id,
                train=self.train,
                encoders=deepcopy(self.encoders),
            )
        else:
            return DataFrame(
                df=self.copy(deep=True),
                slice_id=self.slice_id,
                train=self.train,
                encoders=deepcopy(self.encoders),
            )

    @property
    def _constructor(self):
        """ """
        return DataFrame


def link(df, ktdf):
    """

    Args:
      df: 
      ktdf: 

    Returns:

    """
    assert isinstance(
        ktdf, DataFrame), "Second dataframe should be of type KTDF, not pd.DF"
    return DataFrame(df=df,
                     train=ktdf.train,
                     encoders=ktdf.encoders,
                     slice_id=ktdf.slice_id)



from IPython.display import display

from kts import config
from kts.core import dataframe
from kts.core.backend.memory import cache
from kts.core.feature_constructor import FeatureConstructor
from kts.util import source_utils


def preview(df, sizes=(2, 4, 6)):
    """Applies function to heads of particular dataframe.
    
    Example:
        .. code::

            @preview(df, sizes=[5, 15])
            def make_ohe_pclass(df):
                ...


    Args:
        df:
        sizes:  (Default value = (2, 4, 6):

    Returns:

    """
    def __preview(function):
        config.IS_PREVIEW_CALL = 1
        try:
            for sz in sizes:
                if isinstance(df, dataframe.DataFrame):
                    ktdf = dataframe.DataFrame(df=df.head(sz),
                                               train=df.train,
                                               encoders=df.encoders)
                else:
                    ktdf = dataframe.DataFrame(df=df.head(sz),
                                               train=True,
                                               encoders={})
                # ktdf = dataframe.DataFrame(df.head(sz), True, {})
                display(function(ktdf))
        except Exception as e:
            config.IS_PREVIEW_CALL = 0
            raise e
        config.IS_PREVIEW_CALL = 0

    return __preview


def register(*args, cache_default=True):
    """Registers function for further caching its calls and restoring source.
    
    Example:
        .. code::

            @register
            def make_ohe_pclass(df):
                ...


    Args:
        *args:
        cache_default:  (Default value = True)

    Returns:

    """
    def __register(func):
        # if source_utils.source_is_saved(func) and not source_utils.matches_cache(func):
        if (func.__name__ + "_fc" in cache.cached_objs()
                and source_utils.get_source(func) !=
                cache.load_obj(func.__name__ + "_fc").source):
            raise NameError(
                "A function with the same name is already registered")

        if func.__name__ + "_fc" in cache.cached_objs():
            return cache.load_obj(func.__name__ + "_fc")
        else:
            functor = FeatureConstructor(func, cache_default)
            cache.cache_obj(functor, functor.__name__ + "_fc")
            return functor

    if args:
        function = args[0]
        return __register(function)
    else:
        return __register


def deregister(name, force=False):
    """Deletes sources and cached calls of a certain function.

    Example:
        .. code::

            deregister('make_new_features')

            deregister('make_new_features', force=True)


    Args:
      name:
      force:  (Default value = False)

    Returns:

    """
    confirmation = ""
    fc_name = name + "_fc"
    df_names = [
        df_name for df_name in cache.cached_dfs()
        if df_name.startswith(name + "__")
    ]

    if not force:
        print("Are you sure you want to delete all these cached files?")
        if fc_name in cache.cached_objs():
            print(fc_name)
        for df_name in df_names:
            print(df_name)
        print("To confirm please print full name of the function:")
        confirmation = input()
    if not force and confirmation != name:
        print("Doesn't match")
        return

    if fc_name in cache.cached_objs():
        print(f"removing {fc_name}")
        cache.remove_obj(fc_name)
    for df_name in df_names:
        print(f"removing {df_name}")
        cache.remove_df(df_name)


def dropper(function):
    """Registers function that won't be cached.
    Is recommended to be used only with functions which actually drop columns or rows and don't produce any new data.
    
    Example:
        .. code:: python

            @dropper
            def drop_pclass(df):
                return stl.column_dropper(['Pclass'])(df)


    Args:
        function:

    Returns:
    """
    deregister(function.__name__, force=True)
    return register(function, cache_default=False)


def selector(function):
    """Registers function that won't be cached.
    Is recommended to be used only with functions which actually select columns or rows and don't produce any new data.
    
    Example:
        .. code::

            @selector
            def select_pclass_cabin(df):
               return stl.column_selector(['Pclass', 'Cabin'])(df)

    Args:
        function:

    Returns:

    """
    deregister(function.__name__, force=True)
    return register(function, cache_default=False)


def helper(func):
    """Save function as helper to store its source
    and be able to define it in any notebook with kts.helpers.define_in_scope()

    Args:
        func: function

    Returns:
        function with .source method

    """
    assert "__name__" in dir(func), "Helper should have a name"
    func.source = source_utils.get_source(func)
    if func.__name__ + "_helper" in cache.cached_objs():
        cache.remove_obj(func.__name__ + "_helper")
    cache.cache_obj(func, func.__name__ + "_helper")
    return func



import datetime
import os
from glob import glob

import pandas as pd

from kts.core.dataframe import DataFrame as KTDF
from kts.util import cache_utils
from kts import config


def allow_service(name):
    """

    Args:
      name: 

    Returns:

    """
    return name in config.SERVICE_DF_NAMES


def allow_all(name):
    """

    Args:
      name: 

    Returns:

    """
    return True


if config.CACHE_POLICY == "service":
    gate = allow_service
elif config.CACHE_POLICY == "everything":
    gate = allow_all
else:
    raise UserWarning(
        f'config.CACHE_POLICY should be either "service" or "everything". '
        f'Now it is "{config.CACHE_POLICY}"')


class Cache:
    """Default LRU cache for DataFrames and objects. Uses both RAM and disk space."""
    def __init__(self):
        self.memory = dict()
        self.last_used = dict()
        self.edited_at = dict()
        self.current_volume = 0

    @staticmethod
    def set_memory_limit(volume):
        """Sets a new memory limit in bytes

        Args:
          volume: new memory limit

        Returns:

        """
        config.MEMORY_LIMIT = volume

    def __release_volume(self, df):
        """
        Removes most unpopular dataframes until it is possible to cache given one
        :param df: dataframe
        :return:
        """

        items = sorted([(time, key) for (key, time) in self.last_used.items()])
        cur = 0
        while self.current_volume + cache_utils.get_df_volume(
                df) > config.MEMORY_LIMIT:
            key = items[cur][1]
            cur += 1
            self.current_volume -= cache_utils.get_df_volume(self.memory[key])
            self.memory.pop(key)
            self.last_used.pop(key)
            self.edited_at.pop(key)

    def is_cached_df(self, name):
        """Checks whether given df is cached

        Args:
          name: name of dataframe

        Returns:
          True or False (cache hit or miss)

        """
        # dict_name = name + '_df'
        # return dict_name in self.memory or os.path.exists(cache_utils.get_path_df(name))
        return os.path.exists(cache_utils.get_path_df(name))

    def cache_df(self, df, name):
        """Caches dataframe with given name

        Args:
          df: df
          name: df name

        Returns:

        """
        if not gate(name):
            return
        if self.is_cached_df(name):
            return
        if cache_utils.get_df_volume(df) > config.MEMORY_LIMIT:
            raise MemoryError

        dict_name = name + "_df"
        self.__release_volume(df)
        cache_utils.save_df(df, cache_utils.get_path_df(name))
        self.memory[dict_name] = df
        self.current_volume += cache_utils.get_df_volume(df)
        self.last_used[dict_name] = datetime.datetime.now()
        self.edited_at[dict_name] = cache_utils.get_time(
            cache_utils.get_path_df(name))

    def load_df(self, name):
        """Loads dataframe from cache

        Args:
          name: name of df

        Returns:

        """
        if not self.is_cached_df(name):
            raise KeyError("No such df in cache")

        dict_name = name + "_df"
        self.last_used[dict_name] = datetime.datetime.now()
        if dict_name in self.memory:
            if self.edited_at[dict_name] != cache_utils.get_time(
                    cache_utils.get_path_df(name)):
                self.memory[dict_name] = cache_utils.load_df(
                    cache_utils.get_path_df(name))
                self.edited_at[dict_name] = cache_utils.get_time(
                    cache_utils.get_path_df(name))
            return self.memory[dict_name]
        else:
            tmp = cache_utils.load_df(cache_utils.get_path_df(name))
            self.__release_volume(tmp)
            self.memory[dict_name] = tmp
            self.edited_at[dict_name] = cache_utils.get_time(
                cache_utils.get_path_df(name))
            self.current_volume += cache_utils.get_df_volume(tmp)
            return tmp

    def remove_df(self, name):
        """Removes dataframe from cache

        Args:
          name: name of df

        Returns:

        """
        dict_name = name + "_df"
        if dict_name in self.memory:
            self.current_volume -= cache_utils.get_df_volume(
                self.memory[dict_name])
            self.memory.pop(dict_name)
        if dict_name in self.last_used:
            self.last_used.pop(dict_name)
        if dict_name in self.edited_at:
            self.edited_at.pop(dict_name)
        if os.path.exists(cache_utils.get_path_df(name)):
            os.remove(cache_utils.get_path_df(name))

    @staticmethod
    def cached_dfs():
        """Returns list of cached dataframes
        :return:

        Args:

        Returns:

        """
        return [
            df.split("/")[-1][:-3]
            for df in sorted(glob(config.STORAGE_PATH + "*" + "_df"),
                             key=os.path.getmtime)
        ]

    def is_cached_obj(self, name):
        """Checks whether object is in cache

        Args:
          name: name of object

        Returns:
          True or False (cache hit or miss)

        """
        # dict_name = name + '_obj'
        # return dict_name in self.memory or os.path.exists(cache_utils.get_path_obj(name))
        return os.path.exists(cache_utils.get_path_obj(name))

    def cache_obj(self, obj, name):
        """Caches object with given name

        Args:
          obj: object
          name: object name

        Returns:

        """
        if self.is_cached_obj(name):
            return

        dict_name = name + "_obj"
        self.memory[dict_name] = obj
        cache_utils.save_obj(obj, cache_utils.get_path_obj(name))
        self.edited_at[dict_name] = cache_utils.get_time(
            cache_utils.get_path_obj(name))

    def load_obj(self, name):
        """Loads object from cache

        Args:
          name: name of object

        Returns:

        """
        if not self.is_cached_obj(name):
            raise KeyError("No such object in cache")

        dict_name = name + "_obj"
        if dict_name in self.memory:
            if self.edited_at[dict_name] != cache_utils.get_time(
                    cache_utils.get_path_obj(name)):
                self.memory[dict_name] = cache_utils.load_obj(
                    cache_utils.get_path_obj(name))
                self.edited_at[dict_name] = cache_utils.get_time(
                    cache_utils.get_path_obj(name))
            return self.memory[dict_name]
        else:
            tmp = cache_utils.load_obj(cache_utils.get_path_obj(name))
            self.memory[dict_name] = tmp
            self.edited_at[dict_name] = cache_utils.get_time(
                cache_utils.get_path_obj(name))
            return tmp

    def remove_obj(self, name):
        """Removes object from cache

        Args:
          name: name of object

        Returns:

        """
        dict_name = name + "_obj"
        if dict_name in self.memory:
            self.memory.pop(dict_name)
        if dict_name in self.last_used:
            self.last_used.pop(dict_name)
        if dict_name in self.edited_at:
            self.edited_at.pop(dict_name)
        if os.path.exists(cache_utils.get_path_obj(name)):
            os.remove(cache_utils.get_path_obj(name))

    @staticmethod
    def cached_objs():
        """Returns list of cached objects
        :return:

        Args:

        Returns:

        """
        return [
            df.split("/")[-1][:-4]
            for df in sorted(glob(config.STORAGE_PATH + "*" + "_obj"),
                             key=os.path.getmtime)
        ]


class RAMCache:
    """LRU cache for DataFrames and objects. Uses only RAM, no disk space is consumed."""
    def __init__(self):
        self.memory = dict()
        self.last_used = dict()
        self.current_volume = 0

    @staticmethod
    def set_memory_limit(volume):
        """Sets a new memory limit in bytes

        Args:
          volume: new memory limit

        Returns:

        """
        config.MEMORY_LIMIT = volume

    def __release_volume(self, df):
        """
        Removes most unpopular dataframes until it is possible to cache given one
        :param df: dataframe
        :return:
        """

        items = sorted([(time, key) for (key, time) in self.last_used.items()])
        cur = 0
        while self.current_volume + cache_utils.get_df_volume(df) > config.MEMORY_LIMIT:
            key = items[cur][1]
            cur += 1
            self.current_volume -= cache_utils.get_df_volume(self.memory[key])
            self.memory.pop(key)
            self.last_used.pop(key)

    def is_cached_df(self, name):
        """Checks whether given df is cached

        Args:
          name: name of dataframe

        Returns:
          True or False (cache hit or miss)

        """
        dict_name = name + "_df"
        return dict_name in self.memory

    def cache_df(self, df, name):
        """Caches dataframe with given name

        Args:
          df: df
          name: df name

        Returns:

        """
        if not gate(name):
            return
        if self.is_cached_df(name):
            return
        if cache_utils.get_df_volume(df) > config.MEMORY_LIMIT:
            raise MemoryError

        dict_name = name + "_df"
        self.__release_volume(df)
        self.memory[dict_name] = df
        self.current_volume += cache_utils.get_df_volume(df)
        self.last_used[dict_name] = datetime.datetime.now()

    def load_df(self, name):
        """Loads dataframe from cache

        Args:
          name: name of df

        Returns:

        """
        if not self.is_cached_df(name):
            raise KeyError("No such df in cache")

        dict_name = name + "_df"
        self.last_used[dict_name] = datetime.datetime.now()
        return self.memory[dict_name]

    def remove_df(self, name):
        """Removes dataframe from cache

        Args:
          name: name of df

        Returns:

        """
        dict_name = name + "_df"
        if dict_name in self.memory:
            self.current_volume -= cache_utils.get_df_volume(
                self.memory[dict_name])
            self.memory.pop(dict_name)
        if dict_name in self.last_used:
            self.last_used.pop(dict_name)

    def cached_dfs(self):
        """Returns list of cached dataframes
        :return:

        Args:

        Returns:

        """
        return [i[:-3] for i in list(self.memory.keys()) if i.endswith("_df")]

    def is_cached_obj(self, name):
        """Checks whether object is in cache

        Args:
          name: name of object

        Returns:
          True or False (cache hit or miss)

        """
        dict_name = name + "_obj"
        return dict_name in self.memory

    def cache_obj(self, obj, name):
        """Caches object with given name

        Args:
          obj: object
          name: object name

        Returns:

        """
        if self.is_cached_obj(name):
            return

        dict_name = name + "_obj"
        self.memory[dict_name] = obj

    def load_obj(self, name):
        """Loads object from cache

        Args:
          name: name of object

        Returns:

        """
        if not self.is_cached_obj(name):
            raise KeyError("No such object in cache")

        dict_name = name + "_obj"
        return self.memory[dict_name]

    def remove_obj(self, name):
        """Removes object from cache

        Args:
          name: name of object

        Returns:

        """
        dict_name = name + "_obj"
        if dict_name in self.memory:
            self.memory.pop(dict_name)

    def cached_objs(self):
        """Returns list of cached objects
        :return:

        Args:

        Returns:

        """
        return [i[:-4] for i in list(self.memory.keys()) if i.endswith("_obj")]


class DiskCache:
    """Saves and loads directly from disk, no RAM boosting."""
    def __init__(self):
        pass

    def is_cached_df(self, name):
        """Checks whether given df is cached

        Args:
          name: name of dataframe

        Returns:
          True or False (cache hit or miss)

        """
        return os.path.exists(cache_utils.get_path_df(name))

    def cache_df(self, df, name):
        """Caches dataframe with given name

        Args:
          df: df
          name: df name

        Returns:

        """
        if not gate(name):
            return
        cache_utils.save_df(df, cache_utils.get_path_df(name))

    def load_df(self, name):
        """Loads dataframe from cache

        Args:
          name: name of df

        Returns:

        """
        if not self.is_cached_df(name):
            raise KeyError("No such df in cache")

        return cache_utils.load_df(cache_utils.get_path_df(name))

    def remove_df(self, name):
        """Removes dataframe from cache

        Args:
          name: name of df

        Returns:

        """
        if os.path.exists(cache_utils.get_path_df(name)):
            os.remove(cache_utils.get_path_df(name))

    @staticmethod
    def cached_dfs():
        """Returns list of cached dataframes
        :return:

        Args:

        Returns:

        """
        return [
            df.split("/")[-1][:-3]
            for df in sorted(glob(config.STORAGE_PATH + "*" + "_df"),
                             key=os.path.getmtime)
        ]

    def is_cached_obj(self, name):
        """Checks whether object is in cache

        Args:
          name: name of object

        Returns:
          True or False (cache hit or miss)

        """
        return os.path.exists(cache_utils.get_path_obj(name))

    def cache_obj(self, obj, name):
        """Caches object with given name

        Args:
          obj: object
          name: object name

        Returns:

        """
        if self.is_cached_obj(name):
            return

        cache_utils.save_obj(obj, cache_utils.get_path_obj(name))

    def load_obj(self, name):
        """Loads object from cache

        Args:
          name: name of object

        Returns:

        """
        if not self.is_cached_obj(name):
            raise KeyError("No such object in cache")

        return cache_utils.load_obj(cache_utils.get_path_obj(name))

    def remove_obj(self, name):
        """Removes object from cache

        Args:
          name: name of object

        Returns:

        """
        if os.path.exists(cache_utils.get_path_obj(name)):
            os.remove(cache_utils.get_path_obj(name))

    @staticmethod
    def cached_objs():
        """Returns list of cached objects
        :return:

        Args:

        Returns:

        """
        return [
            df.split("/")[-1][:-4]
            for df in sorted(glob(config.STORAGE_PATH + "*" + "_obj"),
                             key=os.path.getmtime)
        ]


if config.CACHE_MODE == "disk_and_ram":
    cache = Cache()
elif config.CACHE_MODE == "ram":
    cache = RAMCache()
elif config.CACHE_MODE == "disk":
    cache = DiskCache()
else:
    raise UserWarning(
        f'config.CACHE_MODE should be one of "disk", "disk_and_ram", "ram". '
        f'Now it is "{config.CACHE_MODE}"')

USER_SEP = "__USER__"


def save(obj, name):
    """

    Args:
      obj: 
      name: 

    Returns:

    """
    if name in ls():
        raise KeyError(
            "You've already saved object with this name. If you want to overwrite it, first use kts.rm()"
        )
    if isinstance(obj, KTDF) or isinstance(obj, pd.DataFrame):
        cache.cache_df(obj, name + USER_SEP)
    else:
        cache.cache_obj(obj, name + USER_SEP)


def ls():
    """ """
    return [i[:i.find(USER_SEP)] for i in cache.cached_dfs() + cache.cached_objs() if USER_SEP in i]


def get_type(name):
    """

    Args:
      name: 

    Returns:

    """
    if cache.is_cached_obj(name + USER_SEP):
        return "obj"
    else:
        return "df"


def load(name):
    """

    Args:
      name: 

    Returns:

    """
    if name not in ls():
        raise KeyError("No such object in cache")
    if get_type(name) == "df":
        return cache.load_df(name + USER_SEP)
    else:
        return cache.load_obj(name + USER_SEP)


def remove(name):
    """

    Args:
      name: 

    Returns:

    """
    if name not in ls():
        return
    if get_type(name) == "df":
        cache.remove_df(name + USER_SEP)
    else:
        cache.remove_obj(name + USER_SEP)


rm = remove



import inspect
from typing import Optional, Union, List, Tuple, Dict, Any

import numpy as np
import pandas as pd

import kts.core.base_constructors
from kts.core import dataframe
from kts.core.backend import memory
from kts.core.feature_constructor import FeatureConstructor
from kts.feature_selection.selector import BuiltinImportance
from kts.util import source_utils, cache_utils

FeatureListType = Union[FeatureConstructor, List[FeatureConstructor],
                        Tuple[FeatureConstructor], List["FeatureListType"],
                        Tuple["FeatureListType"], ]


class FeatureSet:
    """ """
    def __init__(
            self,
            fc_before: FeatureListType,
            fc_after: FeatureListType = kts.core.base_constructors.empty_like,
            df_input: Optional[pd.DataFrame] = None,
            target_columns: Optional[Union[List[str], str]] = None,
            auxiliary_columns: Optional[Union[List[str], str]] = None,
            name: Optional[str] = None,
            description: Optional[str] = None,
            desc: Optional[str] = None,
            encoders: Optional[Dict[str, Any]] = None,
    ):
        if type(fc_before) == list:
            self.fc_before = kts.core.base_constructors.concat(fc_before)
        elif type(fc_before) == tuple:
            self.fc_before = kts.core.base_constructors.compose(fc_before)
        else:
            self.fc_before = fc_before
        if type(fc_after) == list:
            self.fc_after = kts.core.base_constructors.concat(fc_after)
        elif type(fc_after) == tuple:
            self.fc_after = kts.core.base_constructors.compose(fc_after)
        else:
            self.fc_after = fc_after
        self.target_columns = target_columns
        if isinstance(self.target_columns, str):
            self.target_columns = [self.target_columns]
        elif self.target_columns is None:
            self.target_columns = []
        self.auxiliary_columns = auxiliary_columns
        if isinstance(self.auxiliary_columns, str):
            self.auxiliary_columns = [self.auxiliary_columns]
        elif self.auxiliary_columns is None:
            self.auxiliary_columns = []
        self.encoders = encoders if encoders is not None else dict()
        if df_input is not None:
            self.set_df(df_input)
        self._first_name = name
        self.__doc__ = None
        if desc is not None and description is not None:
            raise ValueError(
                "desc is an alias of description. You can't use both")
        if description is not None:
            self.__doc__ = description
        elif desc is not None:
            self.__doc__ = desc
        self.altsource = None

    def set_df(self, df_input):
        """

        Args:
          df_input: 

        Returns:

        """
        self.df_input = dataframe.DataFrame(df=df_input)
        self.df_input.train = True
        self.df_input.encoders = self.encoders
        self.df = self.fc_before(self.df_input)

    def __call__(self, df):
        ktdf = dataframe.DataFrame(df=df)
        ktdf.encoders = self.encoders
        return kts.core.base_constructors.merge([self.fc_before(ktdf), self.fc_after(ktdf)])

    def __getitem__(self, idx):
        if isinstance(self.df_input, type(None)):
            raise AttributeError("Input DataFrame is not defined")
        return kts.core.base_constructors.merge([
            self.df.iloc[idx],
            self.fc_after(
                dataframe.DataFrame(df=self.df_input.iloc[idx],
                                    train=1)),  # BUG: should have .train=True?
        ])  # made .train=1 only for preview purposes
        # actually, FS[a:b] functionality is made only for debug
        # why not write config.IS_PREVIEW_CALL = 1 then?

    def empty_copy(self):
        """ """
        res = FeatureSet(
            fc_before=self.fc_before,
            fc_after=self.fc_after,
            target_columns=self.target_columns,
            auxiliary_columns=self.auxiliary_columns,
            encoders=self.encoders,
            name=self.__name__,
            description=self.__doc__,
        )
        res.altsource = self.altsource
        return res

    def slice(self, idx_train, idx_test):
        """

        Args:
          idx_train: 
          idx_test: 

        Returns:

        """
        return FeatureSlice(self, idx_train, idx_test)

    @property
    def target(self):
        """ """
        if len(self.target_columns) > 0:
            if set(self.target_columns) < set(self.df_input.columns):
                return self.df_input[self.target_columns]
            elif set(self.target_columns) < set(self.df_input.columns):
                return self.df[self.target_columns]
            else:
                raise AttributeError("Target columns are neither given as input nor computed")
        else:
            raise AttributeError("Target columns are not defined.")

    @property
    def auxiliary(self):
        """ """
        if len(self.auxiliary_columns) > 0:
            if set(self.auxiliary_columns) < set(self.df_input.columns):
                return self.df_input[self.auxiliary_columns]
            elif set(self.auxiliary_columns) < set(self.df_input.columns):
                return self.df[self.auxiliary_columns]
            else:
                raise AttributeError("Auxiliary columns are neither given as input nor computed")
        else:
            raise AttributeError("Auxiliary columns are not defined.")

    @property
    def aux(self):
        """ """
        return self.auxiliary

    def __get_src(self, fc):
        if fc.__name__ == "empty_like":
            return "stl.empty_like"
        if not isinstance(fc, FeatureConstructor):
            return source_utils.get_source(fc)
        if fc.stl:
            return fc.source
        else:
            return fc.__name__

    def _source(self, short=True):
        """

        Args:
          short:  (Default value = True)

        Returns:

        """
        fc_before_source = self.__get_src(self.fc_before)
        fc_after_source = self.__get_src(self.fc_after)
        if short:
            fc_before_source = source_utils.shorten(fc_before_source)
            fc_after_source = source_utils.shorten(fc_after_source)
        prefix = "FeatureSet("
        fs_source = (prefix + "fc_before=" + fc_before_source + ",\n" +
                     " " * len(prefix) + "fc_after=" + fc_after_source +
                     ",\n" + " " * len(prefix) + "target_columns=" +
                     repr(self.target_columns) + ", auxiliary_columns=" +
                     repr(self.auxiliary_columns) + ")")
        return fs_source

    @property
    def source(self):
        """ """
        return self._source(short=False)

    def recover_name(self):
        """ """
        if self._first_name:
            return self._first_name
        ans = []
        frame = inspect.currentframe()
        tmp = {}
        for i in range(7):
            try:
                tmp = {**tmp, **frame.f_globals}
            except:
                pass
            try:
                frame = frame.f_back
            except:
                pass

        for k, var in tmp.items():
            if isinstance(var, self.__class__):
                if hash(self) == hash(var) and k[0] != "_":
                    ans.append(k)
        if len(ans) != 1:
            print(f"The name cannot be uniquely defined. Possible options are: {ans}. "
                  f"Choosing {ans[0]}. You can set the name manually via FeatureSet(name=...) or using .set_name(...)")
        self._first_name = ans[0]
        return self._first_name

    @property
    def __name__(self):
        return self.recover_name()

    def set_name(self, name):
        """

        Args:
          name: 

        Returns:

        """
        self._first_name = name

    def __repr__(self):
        if "altsource" in self.__dict__ and self.altsource is not None:
            return self.altsource
        else:
            return self._source(short=True)

    def select(self,
               n_best,
               experiment,
               calculator=BuiltinImportance(),
               mode="max"):
        """

        Args:
          n_best: 
          experiment: 
          calculator:  (Default value = BuiltinImportance())
          mode:  (Default value = 'max')

        Returns:

        """
        good_features = list(
            experiment.feature_importances(
                importance_calculator=calculator).agg(mode).sort_values(
                    ascending=False).head(n_best).index)
        res = FeatureSet(
            fc_before=self.fc_before + good_features,
            fc_after=self.fc_after + good_features,
            df_input=self.df_input,
            target_columns=self.target_columns,
            auxiliary_columns=self.auxiliary_columns,
            encoders=self.encoders,
            name=f"{self.__name__}_{calculator.short_name}_{n_best}",
            description=
            f"Select {n_best} best features from {self._first_name} using {calculator.__class__.__name__}",
        )
        res.altsource = (
            self.__repr__() +
            f".select({n_best}, lb['{experiment.identifier}'], {calculator.__repr__()})"
        )
        return res


class FeatureSlice:
    """ """
    def __init__(self, featureset, slice, idx_test):
        self.featureset = featureset
        self.slice = slice
        self.idx_test = idx_test
        self.slice_id = cache_utils.get_hash_slice(slice)
        self.first_level_encoders = self.featureset.encoders
        self.second_level_encoders = {}
        self.columns = None
        # self.df_input = copy(self.featureset.df_input)

    def __call__(self, df=None):
        if isinstance(df, type(None)):
            fsl_level_df = dataframe.DataFrame(
                df=self.featureset.df_input.iloc[self.slice],
                # ALERT: may face memory leak here
                slice_id=self.slice_id,
                train=True,
                encoders=self.second_level_encoders,
            )
            result = kts.core.base_constructors.merge([
                self.featureset.df.iloc[self.slice],
                self.featureset.fc_after(fsl_level_df),
            ])
            self.columns = [
                i for i in result.columns
                if i not in self.featureset.target_columns
                and i not in self.featureset.auxiliary_columns
            ]
            return result[self.columns]
        elif (isinstance(df, slice) or isinstance(df, np.ndarray)
              or isinstance(df, list)):
            fsl_level_df = dataframe.DataFrame(
                df=self.featureset.df_input.
                iloc[df],  # ALERT: may face memory leak here
                slice_id=self.slice_id,
                train=False,
                encoders=self.second_level_encoders,
            )
            result = kts.core.base_constructors.merge([
                self.featureset.df.iloc[df],
                self.featureset.fc_after(fsl_level_df)
            ])
            for column in set(self.columns) - set(result.columns):
                result[column] = 0
            return result[self.columns]
        else:
            fs_level_df = dataframe.DataFrame(
                df=df, encoders=self.first_level_encoders)
            fsl_level_df = dataframe.DataFrame(
                df=df,
                encoders=self.second_level_encoders,
                slice_id=self.slice_id)
            result = kts.core.base_constructors.merge([
                self.featureset.fc_before(
                    fs_level_df),  # uses FeatureSet-level encoders
                self.featureset.fc_after(
                    fsl_level_df),  # uses FeatureSlice-level encoders
            ])
            for column in set(self.columns) - set(result.columns):
                result[column] = 0
            return result[self.columns]

    @property
    def target(self):
        """ """
        return self.featureset.target.iloc[self.slice]

    def compress(self):
        """ """
        self.featureset = self.featureset.empty_copy()


from collections import MutableSequence


class FeatureList(MutableSequence):
    """ """
    def __init__(self):
        self.full_name = "kts.api.feature.feature_list"  # such a hardcode
        self.names = [self.full_name]
        while self.names[-1].count("."):
            self.names.append(self.names[-1][self.names[-1].find(".") + 1:])
        self.names.append("kts.features")
        while self.names[-1].count("."):
            self.names.append(self.names[-1][self.names[-1].find(".") + 1:])
        self.functors = []
        self.name_to_idx = dict()

    def recalc(self):
        """ """
        self.functors = []
        self.name_to_idx = dict()
        names = [
            obj for obj in memory.cache.cached_objs() if obj.endswith("_fc")
        ]
        for idx, name in enumerate(names):
            functor = memory.cache.load_obj(name)
            self.functors.append(functor)
            self.name_to_idx[functor.__name__] = idx

    def __repr__(self):
        self.recalc()
        string = f"[{', '.join([f.__str__() for f in self.functors])}]"
        return string

    def __getitem__(self, key):
        self.recalc()
        if type(key) in [int, slice]:
            return self.functors[key]
        elif type(key) == str:
            return self.functors[self.name_to_idx[key]]
        else:
            raise TypeError("Index should be int, slice or str")

    def __delitem__(self, key):
        raise AttributeError("This object is read-only")

    def __setitem__(self, key, value):
        raise AttributeError("This object is read-only")

    def insert(self, key, value):
        """

        Args:
          key: 
          value: 

        Returns:

        """
        raise AttributeError("This object is read-only")

    def define_in_scope(self, global_scope):
        """

        Args:
          global_scope: 

        Returns:

        """
        self.recalc()
        for func in self.name_to_idx:
            for name in self.names:
                try:
                    exec(f"{func} = {name}['{func}']", global_scope)
                    break
                except BaseException:
                    pass

    def __len__(self):
        self.recalc()
        return len(self.functors)


feature_list = FeatureList()



import os

import click

from kts.cli.utils import check_existence, clear_all, list_files, create_config

VALID_EXAMPLES = ['titanic']
KTS_PROJECT_PATHS = ['./input', './notebooks', './storage', './output', './submissions']


@click.group()
def cli():
    pass


@cli.command()
def init():
    if check_existence(KTS_PROJECT_PATHS):
        list_files('./')
        if click.confirm('Do you want to clear existing kts file system?'):
            clear_all()

    click.confirm('Do you want to build the file system?', abort=True)
    for path in KTS_PROJECT_PATHS:
        if not os.path.isdir(path):
            os.makedirs(path)

    if os.path.exists('./kts_config.py'):
        if click.confirm('Config found. Overwrite?'):
            create_config()


@cli.command()
@click.argument('name', type=click.Choice(VALID_EXAMPLES))
def download_example(name):
    f""""""
    if os.path.exists(name):
        raise OSError(f'Path ./{name} already exists.')
    download_name = ".kts_examples_" + str(int(time.time()))
    os.system(f"git clone https://github.com/konodyuk/kts-examples {download_name}")
    shutil.move(f"{download_name}/{name}", f"{name}")
    os.system(f"""cd {name} && echo "n\ny\ny" | kts init""")
    shutil.rmtree(f"{download_name}")




from kts.core.base_constructors import empty_like, identity, concat, compose, column_selector, column_dropper
from kts.stl.categorical import make_ohe, target_encoding, target_encode_list
from kts.stl.numeric import discretize, discretize_quantile, standardize
from kts.stl.misc import apply, get_categorical, get_numeric, stack, from_df



import copy

import pandas as pd
from sklearn.preprocessing import StandardScaler

from kts.core.base_constructors import empty_like, wrap_stl_function
from kts.stl.misc import get_numeric


def discretize(cols, bins, prefix="disc_"):
    """

    Args:
      cols:
      bins:
      prefix:  (Default value = 'disc_')

    Returns:

    """
    def __discretize(df):
        res = empty_like(df)
        for col in cols:
            if df.train:
                res[prefix + str(bins) + "_" + col], enc = pd.cut(df[col],
                                                                  bins,
                                                                  retbins=True)
                df.encoders[f"__disc_{bins}_{col}"] = enc
            else:
                enc = df.encoders[f"__disc_{bins}_{col}"]
                res[prefix + str(bins) + "_" + col] = pd.cut(df[col], enc)
        return res

    return wrap_stl_function(discretize, __discretize)


def discretize_quantile(cols, bins, prefix="disc_q_"):
    """

    Args:
      cols:
      bins:
      prefix:  (Default value = 'disc_q_')

    Returns:

    """
    def __discretize_quantile(df):
        res = empty_like(df)
        for col in cols:
            if df.train:
                res[prefix + str(bins) + "_" + col], enc = pd.qcut(
                    df[col], bins, retbins=True)
                df.encoders[f"__disc_q_{bins}_{col}"] = enc
            else:
                enc = df.encoders[f"__disc_q_{bins}_{col}"]
                res[prefix + str(bins) + "_" + col] = pd.cut(df[col], enc)
        return res

    return wrap_stl_function(discretize_quantile, __discretize_quantile)


def standardize(cols=None, prefix="std_"):
    """

    Args:
      cols:  (Default value = None)
      prefix:  (Default value = 'std_')

    Returns:

    """
    def __standardize(df):
        res = empty_like(df)
        cols_inner = copy.copy(cols)
        if cols_inner is None:
            cols_inner = get_numeric(df)
        for col in cols_inner:
            if df.train:
                encoder = StandardScaler()
                res[prefix + col] = encoder.fit_transform(df[[col]].values)
                df.encoders[f"__std_{col}"] = encoder
            else:
                encoder = df.encoders[f"__std_{col}"]
                res[prefix + col] = encoder.transform(df[[col]].values)
        return res

    return wrap_stl_function(standardize, __standardize)